# iosdiff.py
#
# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>



"""Cisco IOS configuration differences module.

This module compares two configurations for Cisco IOS devices and
produces a delta configuration.
"""



# --- imports ---



import re

from .configdiff import DiffConvert, DiffConfig, pathstr



# --- functions ---



def _print_list(l, prefix="", start=None, step=10):
    """This function 'prints' a list of strings 'l' by adding a prefix
    onto the start of each entry and numbering them.  It's used to
    render IP access-lists and prefix-lists.

    The 'prefix' is a fixed string and defaults to the empty string.
    This is correct for access-lists but could be set to 'ip prefix-list
    NAME' for a prefix-list.  If this contains text, it will most likely
    need to end with a space to separate it from the rest of the string.

    The 'start' and 'step' arguments control the numbering of the
    entries and are added next.  If a start value is not specified, it
    defaults to the step value; the step value defaults to 10.

    A space and the entry value is then added.

    The resulting 'printed' list is then returned to be added to the
    configuration.
    """

    if start is None:
        start = step

    # initialise the result list
    r = []

    n = start
    for i in l:
        r.append(prefix + str(n) + " " + i)
        n += step

    return r



def _isintphysical(i):
    """This function returns if a particular interface is physical or
    not.  The names it matches are those that the are used as the
    canonical ones in the parser.
    """

    return re.match("^(Mgmt|Eth|Fa|Gi|Te|Fo)\d", i)



# --- converter classes ---



# _cvtrs = []
#
# This is a list of converter classes to be added to the
# CiscoIOSConfigDiff object by the _add_converters() method.  Each
# class will be instantiated and added.

_cvtrs = []



# SYSTEM



class _Cvtr_Hostname(DiffConvert):
    path = "hostname",

    def remove(self, cfg, diff, args):
        return "no hostname"

    def update(self, cfg, diff, args):
        return "hostname " + diff

_cvtrs.append(_Cvtr_Hostname)



# INTERFACE ...



class _Cvtr_Int(DiffConvert):
    path = "interface", None

    def remove(self, cfg, diff, args):
        name, = args

        # only remove the interface if it's not physical
        if _isintphysical(name):
            return

        # remove if interface no longer exists, or exists flag has gone
        if (not diff) or ("exists" in diff):
            return "no interface " + name

    def update(self, cfg, diff, args):
        name, = args
        if "exists" in diff:
            return "interface " + name

_cvtrs.append(_Cvtr_Int)


# we put the 'interface / shutdown' at the start to shut it down before
# we do any [re]configuration

class _Cvtr_Int_Shutdown(DiffConvert):
    path = "interface", None, "shutdown"
    name = "shutdown"

    def update(self, cfg, diff, args):
        name, = args
        return "interface " + name, " shutdown"

_cvtrs.append(_Cvtr_Int_Shutdown)


class _Cvtr_Int_CDPEna(DiffConvert):
    path = "interface", None, "cdp-enable"

    def remove(self, cfg, diff, args):
        name, = args
        # if the 'cdp enable' option is not present, that doesn't mean
        # it's disabled but just that it's not specified, so we assume
        # the default is for it to be enabled
        return "interface " + name, " cdp enable"

    def update(self, cfg, diff, args):
        name, = args
        return "interface " + name, " %scdp enable" % ("" if cfg else "no ")

#_cvtrs.append(_Cvtr_Int_CDPEna)


class _Cvtr_Int_ChnGrp(DiffConvert):
    path = "interface", None, "channel-group"

    def remove(self, cfg, diff, args):
        name, = args
        return "interface " + name, " no channel-group"

    def update(self, cfg, diff, args):
        name, = args
        id_, mode = cfg
        return ("interface " + name, " channel-group %d%s"
                    % (id_, mode if mode else ""))

_cvtrs.append(_Cvtr_Int_ChnGrp)


class _Cvtr_Int_Desc(DiffConvert):
    path = "interface", None, "description"

    def remove(self, cfg, diff, args):
        name, = args
        return "interface " + name, " no description"

    def update(self, cfg, diff, args):
        name, = args
        return "interface " + name, " description " + diff

_cvtrs.append(_Cvtr_Int_Desc)


class _Cvtr_Int_Encap(DiffConvert):
    path = "interface", None, "encap"

    def remove(self, cfg, diff, args):
        name, = args
        return "interface " + name, " no encapsulation " + cfg

    def update(self, cfg, diff, args):
        name, = args
        return "interface " + name, " encapsulation " + diff

_cvtrs.append(_Cvtr_Int_Encap)


class _Cvtr_Int_IPAccGrp(DiffConvert):
    path = "interface", None, "ip-access-group"

    def remove(self, cfg, diff, args):
        name, = args
        l = ["interface " + name]
        # if there are differences, just removed the specific group(s);
        # if the whole block is gone, remove all groups there before
        for dir in sorted(diff if diff else cfg):
            l.append(" no ip access-group " + dir)
        return l

    def update(self, cfg, diff, args):
        name, = args
        l = ["interface " + name]
        for dir in sorted(diff):
            l.append(" ip access-group %s %s" % (diff[dir], dir))
        return l

_cvtrs.append(_Cvtr_Int_IPAccGrp)


class _Cvtr_Int_IPAddr(DiffConvert):
    path = "interface", None, "ip-addr"

    def remove(self, cfg, diff, args):
        name, = args
        return "interface " + name, " no ip address"

    def update(self, cfg, diff, args):
        name, = args
        return "interface " + name, " ip address " + diff

_cvtrs.append(_Cvtr_Int_IPAddr)


class _Cvtr_Int_IPAddrSec(DiffConvert):
    path = "interface", None, "ip-addr-sec"

    def remove(self, cfg, diff, args):
        name, = args
        l = ["interface %s" % name]
        for addr in sorted(diff if diff else cfg):
            l.append(" no ip address %s secondary" % addr)
        return l

    def update(self, cfg, diff, args):
        name, = args
        l = ["interface %s" % name]
        for addr in sorted(diff):
            l.append(" ip address %s secondary" % addr)
        return l

_cvtrs.append(_Cvtr_Int_IPAddrSec)


class _Cvtr_Int_IPHlp(DiffConvert):
    path = "interface", None, "ip-helper-addr"

    def remove(self, cfg, diff, args):
        name, = args
        l = ["interface %s" % name]
        if not diff:
            l.append(" no ip helper-address")
        else:
            l.extend([ (" no ip helper-address " + a) for a in diff ])
        return l

    def update(self, cfg, diff, args):
        name, = args
        l = ["interface %s" % name]
        l.extend([ (" ip helper-address " + a) for a in diff ])
        return l

_cvtrs.append(_Cvtr_Int_IPHlp)


class _Cvtr_Int_IP6Addr(DiffConvert):
    path = "interface", None, "ipv6-address"

    def remove(self, cfg, diff, args):
        name, = args
        l = ["interface %s" % name]
        for addr in sorted(diff if diff else cfg):
            l.append(" no ipv6 address " + addr)
        return l

    def update(self, cfg, diff, args):
        name, = args
        l = ["interface %s" % name]
        for addr in sorted(diff):
            l.append(" ipv6 address " + addr)
        return l

_cvtrs.append(_Cvtr_Int_IP6Addr)


class _Cvtr_Int_ServPol(DiffConvert):
    path = "interface", None, "service-policy"

    def remove(self, cfg, diff, args):
        name, = args
        l = ["interface %s" % name]
        for policy in sorted(diff if diff else cfg):
            l.append(" no service-policy " + policy)
        return l

    def update(self, cfg, diff, args):
        name, = args
        l = ["interface %s" % name]
        for policy in sorted(diff):
            l.append(" service-policy " + policy)
        return l

_cvtrs.append(_Cvtr_Int_ServPol)


class _Cvtr_Int_StandbyIP(DiffConvert):
    path = "interface", None, "standby", None, "ip"

    def remove(self, cfg, diff, args):
        name, grp = args
        return "interface " + name, " no standby %d ip" % grp

    def update(self, cfg, diff, args):
        name, grp = args
        return "interface " + name, " standby %d ip %s" % (grp, diff)

#_cvtrs.append(_Cvtr_Int_StandbyIP)


class _Cvtr_Int_StandbyIPSec(DiffConvert):
    path = "interface", None, "standby", None, "ip-sec"

    def remove(self, cfg, diff, args):
        name, grp = args
        l = ["interface " + name]
        for addr in sorted(diff if diff else cfg):
            l.append(" no standby %d ip %s secondary" % (grp, addr))
        return l

    def update(self, cfg, diff, args):
        name, grp = args
        l = ["interface " + name]
        for addr in sorted(diff):
            l.append(" standby %d ip %s secondary" % (grp, addr))
        return l

#_cvtrs.append(_Cvtr_Int_StandbyIPSec)


class _Cvtr_Int_SwPortTrkNtv(DiffConvert):
    # we just match the interface as we need to look inside it to see if
    # the interface is part of a channel group
    path = "interface", None
    ext = "swport-trunk-native",

    def remove(self, cfg, diff, args):
        name, = args

        # if this interface is in a port-channel, we do all changes
        # there, so skip this
        if "channel-group" in cfg:
            return None

        return "interface " + name, " no switchport trunk native vlan"

    def update(self, cfg, diff, args):
        name, = args

        # if this interface is in a port-channel, we do all changes
        # there, so skip this
        if "channel-group" in cfg:
            return None

        return ("interface " + name,
                " switchport trunk native vlan " + self.get_ext(diff))

_cvtrs.append(_Cvtr_Int_SwPortTrkNtv)


class _Cvtr_Int_SwPortTrkAlw(DiffConvert):
    # we just match the interface as we need to look inside it to see if
    # the interface is part of a channel group
    path = "interface", None
    ext = "swport-trunk-allow",

    def remove(self, cfg, diff, args):
        name, = args

        # if this interface is in a port-channel, we do all changes
        # there, so skip this
        if "channel-group" in cfg:
            return None

        l = ["interface " + name]
        if not self.get_ext(diff):
            l.append(" switchport trunk allowed vlan none")
        else:
            for tag in sorted(self.get_ext(diff)):
                l.append(" switchport trunk allowed vlan remove %d" % tag)
        return l

    def update(self, cfg, diff, args):
        name, = args

        # if this interface is in a port-channel, we do all changes
        # there, so skip this
        if "channel-group" in cfg:
            return None

        l = ["interface " + name]
        for tag in sorted(self.get_ext(diff)):
            l.append(" switchport trunk allowed vlan add %d" % tag)
        return l

_cvtrs.append(_Cvtr_Int_SwPortTrkAlw)


class _Cvtr_Int_VRFFwd(DiffConvert):
    path = "interface", None, "vrf-forwarding"

    def remove(self, cfg, diff, args):
        name, = args
        return "interface " + name, " no vrf forwarding"

    def update(self, cfg, diff, args):
        name, = args
        return "interface " + name, " vrf forwarding " + diff
        # TODO: need to find some way to trigger reapplication of IP
        # information (address, HSRP, etc.)

_cvtrs.append(_Cvtr_Int_VRFFwd)


class _Cvtr_Int_XConn(DiffConvert):
    path = "interface", None, "xconnect"

    def remove(self, cfg, diff, args):
        name, = args
        return "interface " + name, " no xconnect"

    def update(self, cfg, diff, args):
        name, = args
        return "interface " + name, " xconnect " + diff

_cvtrs.append(_Cvtr_Int_XConn)


# we put the 'interface / no shutdown' at the end to only enable the
# interface once it's been correctly [re]configured

class _Cvtr_Int_NoShutdown(DiffConvert):
    path = "interface", None, "shutdown"
    name = "no-shutdown"

    def remove(self, cfg, diff, args):
        name, = args
        return "interface " + name, " no shutdown"

_cvtrs.append(_Cvtr_Int_NoShutdown)



# IP[V6] ACCESS-LIST ...



class _Cvtr_IPACL_Std(DiffConvert):
    path = "ip-acl-std", None

    def remove(self, cfg, diff, args):
        name, = args
        return (["no ip access-list standard " + name]
                + _print_list(cfg, "!- "))

    def update(self, cfg, diff, args):
        name, = args
        return ["ip access-list standard " + name] + _print_list(cfg, " ")

_cvtrs.append(_Cvtr_IPACL_Std)


class _Cvtr_IPACL_Ext(DiffConvert):
    path = "ip-acl-ext", None

    def remove(self, cfg, diff, args):
        # include the old list commented out, for comparison
        name, = args
        return (["no ip access-list extended " + name]
                + _print_list(cfg, "!- "))

    def update(self, cfg, diff, args):
        name, = args
        return ["ip access-list extended " + name] + _print_list(diff, " ")

_cvtrs.append(_Cvtr_IPACL_Ext)


class _Cvtr_IPV6ACL_Ext(DiffConvert):
    path = "ipv6-acl", None

    def remove(self, cfg, diff, args):
        name, = args
        return (["no ipv6 access-list " + name]
                + _print_list(cfg, "!- "))

    def update(self, cfg, diff, args):
        name, = args
        return ["ipv6 access-list " + name] + _print_list(diff, " ")

_cvtrs.append(_Cvtr_IPV6ACL_Ext)



# IP[V6] PREFIX-LIST ...



class _Cvtr_IPPfxList(DiffConvert):
    path = "ip-pfx", None

    def remove(self, cfg, diff, args):
        # include the old list commented out, for comparison
        name, = args
        return (["no ip prefix-list " + name]
                + _print_list(cfg, "!- ", step=5))

    def update(self, cfg, diff, args):
        name, = args
        return _print_list(diff, "ip prefix-list %s seq " % name, step=5)

_cvtrs.append(_Cvtr_IPPfxList)


class _Cvtr_IPV6PfxList(DiffConvert):
    path = "ipv6-pfx", None

    def remove(self, cfg, diff, args):
        # include the old list commented out, for comparison
        name, = args
        return (["no ipv6 prefix-list " + name]
                + _print_list(cfg, "!- ", step=5))

    def update(self, cfg, diff, args):
        name, = args
        return _print_list(diff, "ipv6 prefix-list %s seq " % name, step=5)

_cvtrs.append(_Cvtr_IPV6PfxList)



# IP[V6] ROUTE ...



class _Cvtr_IPRoute(DiffConvert):
    path = "ip-route", None

    def remove(self, cfg, diff, args):
        route, = args
        return "no ip route " + route

    def update(self, cfg, diff, args):
        route, = args
        return "ip route " + route

_cvtrs.append(_Cvtr_IPRoute)


class _Cvtr_IP6Route(DiffConvert):
    path = "ipv6-route", None

    def remove(self, cfg, diff, args):
        route, = args
        return "no ipv6 route " + route

    def update(self, cfg, diff, args):
        route, = args
        return "ipv6 route " + route

_cvtrs.append(_Cvtr_IP6Route)



# [NO] SPANNING-TREE ...



class _Cvtr_NoSTP(DiffConvert):
    path = "no-stp", None

    def remove(self, cfg, diff, args):
        tag, = args
        # removing 'no spanning-tree' enables spanning-tree
        return "spanning-tree vlan " + str(tag)

    def update(self, cfg, diff, args):
        tag, = args
        # adding 'no spanning-tree' disables spanning-tree
        return "no spanning-tree vlan " + str(tag)

_cvtrs.append(_Cvtr_NoSTP)


class _Cvtr_STP_Pri(DiffConvert):
    path = "stp-priority", None

    def remove(self, cfg, diff, args):
        tag, = args
        return "no spanning-tree vlan %d priority" % tag

    def update(self, cfg, diff, args):
        tag, = args
        return "spanning-tree vlan %d priority %d" % (tag, diff)

_cvtrs.append(_Cvtr_STP_Pri)



# VLAN ...



class _Cvtr_Vlan(DiffConvert):
    path = "vlan", None

    def remove(self, cfg, diff, args):
        tag, = args
        return "no vlan %d" % tag

    def update(self, cfg, diff, args):
        tag, = args
        if diff and ("exists" in diff):
            return "vlan %d" % tag

_cvtrs.append(_Cvtr_Vlan)


class _Cvtr_Vlan_Name(DiffConvert):
    path = "vlan", None, "name"

    def remove(self, cfg, diff, args):
        tag, = args
        return "vlan %d" % tag, " no name"

    def update(self, cfg, diff, args):
        tag, = args
        return "vlan %d" % tag, " name " + diff

_cvtrs.append(_Cvtr_Vlan_Name)



# --- context parser ---



class CiscoIOSDiffConfig(DiffConfig):
    """This class is used to compare two IOS configuration files and
    generate a configuration file to transform one into the other.
    """


    def _add_converters(self):
        "This method adds the converters for Cisco IOS."

        for cvt_class in _cvtrs:
            self._add_converter(cvt_class())


    def _explain_comment(self, path):
        """This method overrides the empty inherited one to return a
        Cisco IOS comment giving the matched path.
        """

        return "! => " + pathstr(path)


    def _changes_end(self):
        """This method overrides the empty inherited one to return a
        single line saying 'end'.
        """

        return ["end"]


    def init_excludes(self):
        """This method extends the inherited one to add some default
        excludes for the default CoPP (Control Plane Policing) IPv4
        extended and IPv6 ACLs.
        """

        super().init_excludes()

        self._excludes.update({
            "ip-acl-ext": {
                "acl-copp-match-igmp": {},
                "acl-copp-match-pim-data": {},
            },
            "ipv6-acl": {
                "acl-copp-match-mld": {},
                "acl-copp-match-ndv6": {},
                "acl-copp-match-ndv6hl": {},
                "acl-copp-match-pimv6-data": {},
            },
        })



# contextdiff.__main__


# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>
#
# This script compares Cisco IOS configuration files and outputs a
# configuration file which will convert the former into the latter.



import net_contextdiff

import argparse
import re
import sys

import yaml

from net_contextdiff.iosparser import CiscoIOSConfig
from net_contextdiff.iosdiff import CiscoIOSDiffConfig



# --- constants ---



# PLATFORMS = dict
#
# This dictionary specifies the available platforms available to compare
# configurations in.  The dictionary is keyed on the user-specified platform
# name and specifies two values in a tuple, giving the object classes to be
# used for that platform: the first is the parser an the second is the
# difference comparator.

PLATFORMS = {
    "ios": (CiscoIOSConfig, CiscoIOSDiffConfig)
}



# --- command line arguments ---



# create the parser and add in the available command line options

parser = argparse.ArgumentParser(
    # override the program name as running this as a __main__ inside a module
    # directory will use '__main__' by default - this name isn't necessarily
    # correct, but it looks better than that
    prog="net-contextdiff",

    # we want the epilog help output to be printed as it and not reformatted or
    # line wrapped
    formatter_class=argparse.RawDescriptionHelpFormatter,

    epilog="""\
Exclude items are added using the -x and -y options: they consist of paths
through the configuration structure with levels separated by colons.  For
example, excluding "interface:Vlan100" will avoid changes to the VLAN 100
interface.

The -e option is useful to add comments to conversions to indicate what path
matched a particular conversion, so excludes can be constructed.

Note that exclude items are case-sensitive and must be matched exactly.
""")


parser.add_argument(
    "-x", "--exclude",
    metavar="PATH",
    dest="exclude_items",
    nargs="+",
    help="add exclude item for parts of the configuration which will "
         "be omitted from the comparison (these will be added to those "
         "read from a file with the -y option, if both are used), "
         "specified as a dictionary path, with keys separated by "
         "colons")

parser.add_argument(
    "-y", "--exclude-filename",
    metavar="FILENAME",
    help="read exclude items from lines in file")

parser.add_argument(
    "-z", "--exclude-yaml-filename",
    metavar="FILENAME",
    help="read exclude items from a YAML file")

parser.add_argument(
    "-Z", "--exclude-yaml-path",
    default="",
    help="path into file specified with -z to where exclude items are "
         "held, specified as a dictionary path, with keys separated by "
         "colons")

parser.add_argument(
    "-e", "--explain",
    action="store_true",
    help="explain which converter paths match a configuration change")

parser.add_argument(
    "-q", "--quiet",
    action="store_true",
    help="when generating configuration for multiple devices, don't "
         "print the name of each device, as it's generated")

parser.add_argument(
    "-O", "--no-output",
    action="store_true",
    help="do not write changes output file - used to test parsing, "
         "exclude and conversion process")

parser.add_argument(
    "-X", "--debug-excludes",
    action="store_true",
    help="dump the tree of items to be excluded from the comparions")

parser.add_argument(
    "-P", "--debug-parser",
    action="store_true",
    help="enable debugging of contextual parsing; assumed by default "
         "if only a 'from' file is specified")

parser.add_argument(
    "-C", "--debug-config",
    action="store_true",
    help="dump the configuration dictionary after parsing")

parser.add_argument(
    "-D", "--debug-diff",
    action="store_true",
    help="dump the difference (removes and updates) between the from "
         "and to configurations")

parser.add_argument(
    "-V", "--debug-convert",
    action="count",
    default=0,
    help="enable debugging of the difference converstion action "
         "processing (multiple levels increase verbosity up to a "
         "maximum of 2)")

parser.add_argument(
    "platform",
    choices=PLATFORMS,
    help="platform used for configuration files")

parser.add_argument(
    "from_filename",
    metavar="from",
    help="initial ('from') configuration file; '%%' can be used to "
         "substitute in the name of the device into the filename")

parser.add_argument(
    "to_filename",
    nargs="?",
    default=None,
    metavar="to",
    help="destination ('to') configuration file; '%%' can be used to "
         "substitute in the name of the device into the filename; if "
         "omitted just parse the file and assume 'debug' mode")

parser.add_argument(
    "output_filename",
    nargs="?",
    metavar="output",
    help="write differences configuration to named file instead of "
         "stdout; '%%' can be used to substitute in the name of the "
         "device into the filename")

parser.add_argument(
    "devicenames",
    metavar="devicename",
    nargs="*",
    help="name(s) of the device(s) to calculate differences in the "
         "configuration for")

parser.add_argument(
    "--version",
    action="version",
    version=("%(prog)s " + net_contextdiff.__version__))


# parse the supplied command line against these options, storing the results

args = parser.parse_args()

config_parser_class, config_diff_class = PLATFORMS[args.platform]

exclude_items = args.exclude_items
exclude_filename = args.exclude_filename
exclude_yaml_filename = args.exclude_yaml_filename
exclude_yaml_path = args.exclude_yaml_path
explain = args.explain
quiet = args.quiet
no_output = args.no_output
debug_excludes = args.debug_excludes
debug_parser = args.debug_parser
debug_config = args.debug_config
debug_diff = args.debug_diff
debug_convert = args.debug_convert
from_filename = args.from_filename
to_filename = args.to_filename
output_filename = args.output_filename
devicenames = args.devicenames


# check a couple of nonsensical configurations aren't being use related
# to multiple devices

if len(devicenames) == 0:
    if from_filename.find("%") != -1:
        print("warning: no device names specified, so operating on a single "
              "file, yet 'from' filename has '%' character - no substitution "
              "will be performed",
              file=sys.stderr)

    if to_filename and (to_filename.find("%") != -1):
        print("warning: no device names specified, so operating on a single "
              "file, yet 'to' filename has '%' character - no substitution "
              "will be performed",
              file=sys.stderr)

    if output_filename and (output_filename.find("%") != -1):
        print("warning: no device names specified, so operating on a single "
              "file, yet 'output' filename has '%' character - no "
              "substitution will be performed",
              file=sys.stderr)


elif len(devicenames) > 1:
    if from_filename.find("%") == -1:
        print("warning: multiple device names specified but 'from' filename "
              "does not contain '%' - same file will be read",
              file=sys.stderr)

    if to_filename and (to_filename.find("%") == -1):
        print("warning: multiple device names specified but 'to' filename "
              "does not contain '%' - same file will be read",
              file=sys.stderr)


    if not output_filename:
        print("warning: multiple device names specified but outputting to "
              "standard output - all configurations will be concatenated",
              file=sys.stderr)

    elif output_filename.find("%") == -1:
        print("error: multiple device names specified but 'output' filename "
              "does not contain '%' - same file would be overwritten",
              file=sys.stderr)

        exit(1)



# --- setup ---



# create the diff object once at the start

diff = config_diff_class(explain, debug_config, debug_diff, debug_convert)


# add any excludes specified on the command line

if exclude_items:
    # add the exclude items specified directly on the command line (do
    # this before reading those from a file, if specified, below)

    diff.add_excludes(exclude_items)


# if an exclude [simple] file was specified, read and add those

if exclude_filename:
    diff.read_excludes(exclude_filename)


# if an exclude YAML file was specified, read that in

if exclude_yaml_filename:
    try:
        diff.read_excludes_yaml(exclude_yaml_filename, exclude_yaml_path)

    except TypeError:
        print("error: exclude YAML path references non-dictionary item "
              "in file", file=sys.stderr)

        exit(1)

    except KeyError:
        print("error: exclude YAML path not found in file", file=sys.stderr)

        exit(1)


if debug_excludes:
    print(">> exclude items:", yaml.dump(diff.get_excludes(),
          default_flow_style=False), sep="\n", file=sys.stderr)



# --- compare ---



def diffconfig(from_filename, to_filename, output_filename=None,
               no_output=False):

    """This function compares the 'from' and 'to' configurations for
    the specified device and writes a difference configuration file
    (one that transforms the configuration of a running device from the
    'from' to the 'to' state).

    The filenames for the configuration files, as well as the output
    are taken from the arguments parsed above and stored in global
    variables used directly by this function.

    The function returns True iff the parsing and comparison succeeded.
    Most problems result in the function aborting with an exception,
    but some minor warnings may be ignored and the program continue
    with devices.
    """

    # this function makes use of global variables defined outside it so
    # must appear here in the code


    # read in the 'from' and 'to' configurations

    from_cfg = config_parser_class(from_filename, debug_parser)

    to_cfg = (config_parser_class(to_filename, debug_parser)
                  if to_filename else None)


    # find the differences

    diffs = diff.convert(from_cfg, to_cfg)


    # if output is suppressed, just stop with success here

    if no_output:
        return True


    # write the differences to a file or standard output

    if output_filename:
        if debug_convert:
            print("debug: writing to output file: %s" % output_filename,
                  file=sys.stderr)

        with open(output_filename, "w") as output_file:
            if diffs:
                print(diffs, file=output_file)

    else:
        if debug_convert:
            print("debug: writing to standard output", file=sys.stderr)

        if diffs:
            print(diffs)


    return True



# this flag will change to False if any configuration fails to generate and
# is used to affect the return code from the script

complete_success = True


if devicenames:
    for devicename in devicenames:
        if not quiet:
            print(devicename)

        complete_success = diffconfig(
                               from_filename.replace("%", devicename),
                               to_filename.replace("%", devicename),
                               output_filename.replace("%", devicename))


else:
    complete_success = diffconfig(
        from_filename, to_filename, output_filename, no_output)


exit(0 if complete_success else 1)



# iosparser.py
#
# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>



"""Cisco IOS configuration module.

This module parses Cisco IOS configuration files into a dictionary.
"""



# --- imports ---



import re

from deepops import deepsetdefault
from netaddr import IPNetwork

from .configparser import IndentedContextualCommand, IndentedContextualConfig



# --- constants ---



# interface names and their canonical representations - the canonical
# versions don't have to be how Cisco IOS always displays them, in any
# particular context, but just so they can consistently matched against
#
# TODO: this (and the calling function) should really be more general
# and handle regular expressions

_INTERFACE_CANONICALS = {
    "ethernet": "Eth",
    "fastethernet": "Fa",
    "gigabitethernet": "Gi",
    "tengigabitethernet": "Te",
    "fortygigabitethernet": "Fo",
    "port-channel": "Po",
    "vl": "Vlan"
}



# _SERVICE_PORTS = dict
#
# Dictionary mapping service names (as displayed/usable in an access-
# list rule) into a port number.
#
# TODO: this list is not complete but parses all the rules we currently
# have in use.  It should be expanded at some point to all services.

_SERVICE_PORTS = {
    "bootps": 67,
    "bootpc": 68,
    "discard": 9,
    "domain": 53,
    "ftp": 21,
    "ftp-data": 20,
    "gopher": 70,
    "ident": 113,
    "isakmp": 500,
    "lpd": 515,
    "mail": 25,
    "non500-isakmp": 4500,
    "ntp": 123,
    "smtp": 25,
    "snmp": 161,
    "tftp": 69,
    "www": 80
}



# --- functions ---



def _expand_set(s):
    """This function exapands a string giving a set of numbers,
    separated by commas, which can include ranges from low to high,
    using hyphens.  The return value will be the set of numbers
    expressed.

    For example, given a string of "1,3-5", a set containing 1, 3, 4
    and 5 will be returned.
    """

    t = set()

    for i in s.split(","):
        i_range = i.split("-")

        if len(i_range) == 1:
            t.add(int(i))
        else:
            t.update(range(int(i_range[0]), int(i_range[1]) + 1))

    return t



def _ipnet4_to_ios(ipnet4):
    """This method converts an IPv4 IPNetwork object into its canonical
    representation in Cisco IOS in a standard access-list.

    Conversions are '0.0.0.0/0' to 'any' and single host networks (i.e.
    simple addresses) into a plain address with no mask.
    """

    if ipnet4 == IPNetwork("0.0.0.0/0"):
        return "any"

    if ipnet4.prefixlen == 32:
        return str(ipnet4.ip)

    return str(ipnet4.ip) + " " + str(ipnet4.hostmask)



def _service_to_port(service):
    """Converts a Cisco service named 'service' (as displayed/usable in
    an access-list rule) into a port number and return it as an
    integer.

    If the service name is not defined, it is assumed to already be
    numeric and is converted to an integer and returned.  If this
    conversion fails, an exception will be raised (whch probably
    indicates a service that is missing from the list).
    """

    return _SERVICE_PORTS.get(service) or int(service)



# IPV4 STANDARD ACCESS CONTROL LIST RULES



# regular expression for matching an IPv4 standard ACL rule, compiled
# once for efficiency

_IPACL_STD_RULE_RE = re.compile(
    r"^"
    r"(?P<action>permit|deny)"
    r" +"

    # we match "0.0.0.0 255.255.255.255" as "any" because the
    # netaddr module assumes the mask is a netmask (= /32) rather
    # than a hostmask, in this case
    r"((?P<any>any|0\.0\.0\.0 255\.255\.255\.255)|"
        r"(?:host )?(?P<host>[0-9.]+)|"
        r"(?P<net>[0-9.]+) (?P<mask>[0-9.]+))"

    r"$")


def _ipacl_std_rule_parse(rule):
    """Parse an IPv4 standard ACL rule, returning a 2-element tuple
    consisting of the action ('permit' or 'deny') and an IPNetwork
    object, specifying the address/network to match.
    """


    match = _IPACL_STD_RULE_RE.match(rule)

    if not match:
        raise ValueError(
            "failed to parse IPv4 standard ACL rule: " + rule)


    # match some special cases and, if not those, match as a
    # network/hostmask

    if match.group("any"):
        ipnet4 = IPNetwork("0.0.0.0/0")

    elif match.group("host"):
        ipnet4 = IPNetwork(match.group("host"))

    else:
        ipnet4 = IPNetwork(match.group("net") + "/" + match.group("mask"))


    return match.group("action"), ipnet4



def _ipacl_std_canonicalize(l):
    """This function sorts an IPv4 standard access-list into a canoncial
    form so two lists can be compared.  The list is supplied and
    returned as a new list of IOS rules.  Note that it is not sorted in
    place), as parsed by _ipacl_std_rule_parse().

    IPv4 standard ACLs are complicated in IOS due to its tendancy to
    reorganise the rules, after they're entered.  It always preserves
    the semantics of the list (never putting an overlapping 'permit' and
    'deny' in the wrong order), but it can move rules which don't
    interact around.  Presumably this is done to optimise processing.

    The solution adopted here is to build the ACLs up in blocks.  Each
    block is a set of rules where the address portions don't overlap;
    these are built, sorting each block into address order, before
    adding the sorted rules in the block into the resulting list.

    The result is lists which are not necessarily in the same order as
    they were constructed, nor how IOS stores them, but two lists should
    at least be in the same order so they can be directly compared.

    This function is applied to each list, after the configuration is
    read, and the returned list used to replace the order in which the
    rules were read.
    """


    # initialise the returned (canonical list)
    result = []

    # initialise the current block of non-overlapping rules
    block = []

    # go through the rules in the supplied list
    for rule in l:
        # parse the rule into action string and IPNetwork
        action, net = _ipacl_std_rule_parse(rule)

        # find out if this rule overlaps with the network of a rule
        # already in the block
        overlap = [ None for _, chk_net
                         in block
                         if ((net.first <= chk_net.last)
                             and (net.last >= chk_net.first)) ]

        if overlap:
            # we had an overlap, add the current block to the returned
            # list and reinitialise it with just this new rule
            result.extend(block)
            block = [(action, net)]

        else:
            # add this rule to the current block and re-sort it on the
            # addresses of the rules in it
            block.append( (action, net) )
            block.sort(key=(lambda rule: rule[1]))

    # we've reached the end, so store the rules in the current block on
    # the end of the list
    result.extend(block)

    # convert the rules back into IOS text format and return them as a
    # list
    return [ (action + " " + _ipnet4_to_ios(net)) for action, net in result ]



# IPV4 EXTENDED ACCESS CONTROL LIST RULES



# regular expression for matching an IPv4 extended ACL rule, compiled
# once for efficiency

_IPACL_EXT_RULE_RE = re.compile(
    r"^"
    r"(?P<action>permit|deny)"
    r" +"
    r"(?P<protocol>ip|icmp|tcp|udp|igmp|pim|gre|esp)"
    r" "
    r"(?P<src_addr>any|host ([0-9.]+)|([0-9.]+) ([0-9.]+))"
    r"( ("
        # 'eq' and 'neq' can support a list of services - we need to match
        # them non-greedy
        r"((?P<src_port_listop>eq|neq) (?P<src_port_list>\S+( \S+)*?))|"

        r"((?P<src_port_1op>lt|gt) (?P<src_port_num>\S+))|"
        r"range (?P<src_port_low>\S+) (?P<src_port_high>\S+)"
    r"))?"
    r" "
    r"(?P<dst_addr>any|host ([0-9.]+)|([0-9.]+) ([0-9.]+))"
    r"( ("
        r"((?P<dst_port_listop>eq|neq) (?P<dst_port_list>\S+( \S+)*?))|"
        r"((?P<dst_port_1op>lt|gt) (?P<dst_port_num>\S+))|"
        r"range (?P<dst_port_low>\S+) (?P<dst_port_high>\S+)|"
        r"(?P<icmp_type>echo(-reply)?)"
    r"))?"
    r"(?P<established> established)?"
    r"(?P<qos> (dscp \S+))?"
    r"(?P<log> (log|log-input))?"
    r"$"
)


def _ipacl_ext_rule_parse(rule):
    """Parse an IPv4 extended ACL rule, returning a 'normalised'
    form of the rule as a string.  The normalised form should allow
    two ACL rules which mean the same thing to be compared using a
    simple string comparison.

    This process mainly involves extracting the port entries and
    [potentially] translating them into port numbers, if they're named
    services (which can be used in rules, plus IOS will translate a
    numbered service to a name, if one matches).

    Note that no attempt is made to check the rule for validity.
    """


    match = _IPACL_EXT_RULE_RE.match(rule)

    if not match:
        raise ValueError(
            "failed to parse IPv4 extended ACL rule: " + rule)


    action, protocol, src_addr, dst_addr = match.group(
        "action", "protocol", "src_addr", "dst_addr")


    # match.group() will return an error if a named group does not exist
    # in the regexp; match.groupdict(), however, will return a default
    # value (None, if not specified) for named groups that do not exist
    #
    # as such, we need to check if the retrieved groups are blank or
    # not, for optional/alternative parts of a rule

    groups = match.groupdict()


    src_port = ""

    if groups["src_port_listop"]:
        # if 'eq' or 'neq' was found for the source port, it will be one
        # or more services, separated by spaces - we need to split the
        # list up and turn each one into a port number, then join the
        # list back together again

        src_port = (
            " %s %s" % (
                groups["src_port_listop"],
                " ".join([str(_service_to_port(s))
                              for s
                              in groups["src_port_list"].split(" ")])))

    elif groups["src_port_1op"]:
        src_port = " %s %d" % (
                       groups["src_port_1op"],
                       _service_to_port(groups["src_port_num"]))

    elif groups["src_port_low"]:
        src_port = " range %d %d" % (
                        _service_to_port(groups["src_port_low"]),
                        _service_to_port(groups["src_port_high"]))


    dst_port = ""

    if groups["dst_port_listop"]:
        dst_port = (
            " %s %s" % (
                groups["dst_port_listop"],
                " ".join([str(_service_to_port(s))
                              for s
                              in groups["dst_port_list"].split(" ")])))

    elif groups["dst_port_1op"]:
        dst_port = " %s %d" % (
                        groups["dst_port_1op"],
                        _service_to_port(groups["dst_port_num"]))

    elif groups["dst_port_low"]:
        dst_port = " range %d %d" % (
                        _service_to_port(groups["dst_port_low"]),
                        _service_to_port(groups["dst_port_high"]))

    # the destination port could also be an ICMP message type
    elif groups["icmp_type"]:
        dst_port = " " + groups["icmp_type"]


    established = groups["established"] or ""

    qos = groups["qos"] or ""

    log = groups["log"] or ""


    return (action + " " + protocol
            + " " + src_addr + src_port
            + " " + dst_addr + dst_port + established + qos + log)



# IPV6 EXTENDED ACCESS CONTROL LIST RULES



# regular expression for matching an IPv6 access-list rule, compiled
# once for efficiency
#
# TODO: we know this doesn't match some of the more complicated rules
# (such as the CP policing ones matching ICMPv6 types) but we're
# excluding those in the output, anyway, so we just ignore them - as
# such, we don't match the end of string ('$')

_ACL6_RULE_RE = re.compile(
    r"^"
    r"(?P<action>permit|deny)"
    r"( "
        r"(?P<protocol>ipv6|icmp|tcp|udp|\d+)"
    r")?"
    r" "
    r"(?P<src_addr>any|host [0-9A-Fa-f:]+|[0-9A-Fa-f:]+/\d+)"
    r"( ("
        r"((?P<src_port_1op>eq|lt|gt|neq) (?P<src_port_num>\S+))|"
        r"range (?P<src_port_low>\S+) (?P<src_port_high>\S+)"
    r"))?"
    r" "
    r"(?P<dst_addr>any|host [0-9A-Fa-f:]+|[0-9A-Fa-f:]+/\d+)"
    r"( ("
        r"((?P<dst_port_1op>eq|lt|gt|neq) (?P<dst_port_num>\S+))|"
        r"range (?P<dst_port_low>\S+) (?P<dst_port_high>\S+)|"
        r"(?P<icmp_type>echo(-reply)?)"
    r"))?"
    r"(?P<established> established)?"
    r"(?P<log> (log|log-input))?"
)


def _acl6_rule_parse(rule):
    """Parse an IPv6 ACL rule, returning a 'normalised' form of the rule
    as a string.  The normalised form should allow two ACL rules which
    mean the same thing to be compared using a simple string comparison.

    This process mainly involves extracting the port entries and
    [potentially] translating them into port numbers, if they're named
    services (which can be used in rules, plus IOS will translate a
    numbered service to a name, if one matches).

    Note that no attempt is made to check the rule for validity.
    """


    match = _ACL6_RULE_RE.match(rule)

    if not match:
        raise ValueError("failed to parse IPv6 ACL rule: " + rule)


    action, protocol, src_addr, dst_addr = match.group(
        "action", "protocol", "src_addr", "dst_addr")


    # if the protocol was not specified, we default to 'ipv6'

    if protocol is None:
        protocol = "ipv6"


    # lower case the source and destination addresses since IPv6
    # addresses can either be in upper or lower case (usually upper, in
    # IOS); we choose lower here, though, to avoid upper-casing the
    # keywords 'host' and 'any'

    src_addr = src_addr.lower()
    dst_addr = dst_addr.lower()


    # match.group() will return an error if a named group does not exist
    # in the regexp; match.groupdict(), however, will return a default
    # value (None, if not specified) for named groups that do not exist
    #
    # as such, we need to check if the retrieved groups are blank or
    # not, for optional/alternative parts of a rule

    groups = match.groupdict()


    src_port = ""

    if groups["src_port_num"]:
        src_port = " %s %d" % (
                       groups["src_port_1op"],
                       _service_to_port(groups["src_port_num"]))

    elif groups["src_port_low"]:
        src_port = " range %d %d" % (
                       _service_to_port(groups["src_port_low"]),
                       _service_to_port(groups["src_port_high"]))


    dst_port = ""

    if groups["dst_port_num"]:
        dst_port = " %s %d" % (
                       groups["dst_port_1op"],
                       _service_to_port(groups["dst_port_num"]))

    elif groups["dst_port_low"]:
        dst_port = " range %d %d" % (
                       _service_to_port(groups["dst_port_low"]),
                       _service_to_port(groups["dst_port_high"]))


    elif groups["icmp_type"]:
        dst_port = " " + groups["icmp_type"]


    established = groups["established"] or ""

    log = groups["log"] or ""


    return (action + " " + protocol
            + " " + src_addr + src_port
            + " " + dst_addr + dst_port + established)



# --- configuration command classes ---



# _cmds = []
#
# This is a list of classes, one for each IOS configuration mode
# command.  The command classes are defined at the global level and
# added to this list.
#
# The CiscoIOSConfig class adds these to the object upon instantiation,
# by the _add_commands() method.
#
# This was done to make it clearer how the commands are implemented.

_cmds = []



# _Cmd is created to be a shorthand for the IndentedContextualCommand
# class as we'll be using it a lot

_Cmd = IndentedContextualCommand



# SYSTEM



class _Cmd_Comment(_Cmd):
    cmd = r"!.*"

_cmds.append(_Cmd_Comment)


class _Cmd_Hostname(_Cmd):
    cmd = r"hostname (?P<hostname>\S+)"

    def action(self, groups, cfg, params):
        cfg["hostname"] = groups["hostname"]

_cmds.append(_Cmd_Hostname)



# INTERFACE ...



class _Cmd_Int(_Cmd):
    cmd = r"interface (?P<name>\S+)"
    enter_context = "interface"

    def action(self, groups, cfg, params):
        name = groups["name"]

        # we need to canonicalise the type of an interface as they can
        # be specified in various shorthand forms, and in mixed case
        #
        # the canonical form is '<type><number>'; the type is first
        # character capitalised
        #
        # TODO: we really should handle this more flexibly, with a
        # central funcation matching regexps of the interface name
        match = re.match(r"^([-a-z]+)([0-9/.]*)$", name.lower())
        if match:
            _type, num = match.groups()
            _type = _INTERFACE_CANONICALS.get(_type, _type).capitalize()
            name = _type + num

        return deepsetdefault(cfg, "interface", name)

_cmds.append(_Cmd_Int)


class _CmdContext_Int(_Cmd):
    context = "interface"


class _Cmd_Int_CDPEna(_CmdContext_Int):
    cmd = r"(?P<no>no )?cdp enable"

    def action(self, groups, cfg, params):
        # we allow CDP to be 'no cdp enable' to clear the CDP status
        cfg["cdp-enable"] = not groups["no"]

_cmds.append(_Cmd_Int_CDPEna)


class _Cmd_Int_ChnGrp(_CmdContext_Int):
    cmd = r"channel-group (?P<id>\d+)(?P<mode> .+)?"

    def action(self, groups, cfg, params):
        cfg["channel-group"] = int(groups["id"]), groups["mode"]

_cmds.append(_Cmd_Int_ChnGrp)


class _Cmd_Int_Desc(_CmdContext_Int):
    cmd = r"description (?P<description>.+)"

    def action(self, groups, cfg, params):
        cfg["description"] = groups["description"]

_cmds.append(_Cmd_Int_Desc)


class _Cmd_Int_Encap(_CmdContext_Int):
    cmd = r"encapsulation (?P<encap>dot1q \d+( native)?)"

    def action(self, groups, cfg, params):
        # lower case the encapsulation definition as IOS stores 'dot1q'
        # as 'dot1Q'
        cfg["encap"] = groups["encap"].lower()

_cmds.append(_Cmd_Int_Encap)


class _Cmd_Int_HlprAddr(_CmdContext_Int):
    cmd = r"ip helper-address (?P<helper_addr>(global )?\S+)"

    def action(self, groups, cfg, params):
        cfg.setdefault("ip-helper-addr", set()).add(groups["helper_addr"])

_cmds.append(_Cmd_Int_HlprAddr)


class _Cmd_Int_IPAccGrp(_CmdContext_Int):
    cmd = r"ip access-group (?P<access_group>\S+) (?P<direction>in|out)"

    def action(self, groups, cfg, params):
        cfg.setdefault("ip-access-group", {})[groups["direction"]] = (
            groups["access_group"])

_cmds.append(_Cmd_Int_IPAccGrp)


class _Cmd_Int_IPAddr(_CmdContext_Int):
    cmd = r"ip address (?P<addr>\S+ \S+)"

    def action(self, groups, cfg, params):
        cfg["ip-addr"] = groups["addr"]

_cmds.append(_Cmd_Int_IPAddr)


class _Cmd_Int_IPAddrSec(_CmdContext_Int):
    cmd = r"ip address (?P<addr>\S+ \S+) secondary"

    def action(self, groups, cfg, params):
        # secondary address - record it in a list
        cfg.setdefault("ip-addr-sec", set()).add(groups["addr"])

_cmds.append(_Cmd_Int_IPAddrSec)


class _Cmd_Int_IP6Addr(_CmdContext_Int):
    cmd = r"ipv6 address (?P<addr>\S+)"

    def action(self, groups, cfg, params):
        # IPv6 addresses involve letters so we lower case for
        # consistency
        cfg.setdefault("ipv6-addr", set()).add(groups["addr"].lower())

_cmds.append(_Cmd_Int_IP6Addr)


class _Cmd_Int_ServPol(_CmdContext_Int):
    cmd = r"service-policy (?P<policy>.+)"

    def action(self, groups, cfg, params):
        cfg.setdefault("service-policy", set()).add(groups["policy"])

_cmds.append(_Cmd_Int_ServPol)


class _Cmd_Int_Shutdown(_CmdContext_Int):
    cmd = r"(?P<no>no )?shutdown"

    def action(self, groups, cfg, params):
        if groups["no"]:
            # we allow interfaces to be 'no shutdown' so have to clear
            # the shutdown option
            cfg.pop("shutdown", None)
        else:
            cfg["shutdown"] = None

_cmds.append(_Cmd_Int_Shutdown)


class _Cmd_Int_StandbyIP(_CmdContext_Int):
    cmd = r"standby (?P<grp>\d+) ip (?P<addr>\S+)"

    def action(self, groups, cfg, params):
        deepsetdefault(
            cfg, "standby", int(groups["grp"]))["ip"] = groups["addr"]

_cmds.append(_Cmd_Int_StandbyIP)


class _Cmd_Int_StandbyIPSec(_CmdContext_Int):
    cmd = r"standby (?P<grp>\d+) ip (?P<addr>\S+) secondary"

    def action(self, groups, cfg, params):
        deepsetdefault(
            cfg, "standby", int(groups["grp"]), "ip-sec", last=set()).add(
                groups["addr"])

_cmds.append(_Cmd_Int_StandbyIPSec)


class _Cmd_Int_SwPortTrkNtv(_CmdContext_Int):
    cmd = r"switchport trunk native vlan (?P<vlan>\d+)"

    def action(self, groups, cfg, params):
        cfg["swport-trunk-native"] = int(groups["vlan"])

_cmds.append(_Cmd_Int_SwPortTrkNtv)


class _Cmd_Int_SwPortTrkAlw(_CmdContext_Int):
    cmd = r"switchport trunk allowed vlan (add )?(?P<vlans>[0-9,-]+)"

    def action(self, groups, cfg, params):
        cfg.setdefault("swport-trunk-allow", set()).update(
            _expand_set(groups["vlans"]))

_cmds.append(_Cmd_Int_SwPortTrkAlw)


class _Cmd_Int_VRF(_CmdContext_Int):
    cmd = (r"vrf forwarding (?P<vrf>\S+)")

    def action(self, groups, cfg, params):
        cfg["vrf-forwarding"] = groups["vrf"]

_cmds.append(_Cmd_Int_VRF)


class _Cmd_Int_XConn(_CmdContext_Int):
    cmd = r"xconnect (?P<remote>[0-9.]+ \d+ .+)"

    def action(self, groups, cfg, params):
        cfg["xconnect"] = groups["remote"]

_cmds.append(_Cmd_Int_XConn)



# IP ACCESS-LIST STANDARD



class _Cmd_ACLStdRule(_Cmd):
    cmd = r"access-list (?P<num>\d{1,2}|1[3-9]\d{2}) (?P<rule>.+)"

    def action(self, groups, cfg, params):
        deepsetdefault(cfg, "ip-acl-std", groups["num"], last=[]).append(
            groups["rule"])

_cmds.append(_Cmd_ACLStdRule)


class _Cmd_IPACL_Std(_Cmd):
    cmd = r"ip access-list standard (?P<name>.+)"
    enter_context = "ip-acl-std"

    def action(self, groups, cfg, params):
        return deepsetdefault(cfg, "ip-acl-std", groups["name"], last=[])

_cmds.append(_Cmd_IPACL_Std)


class _Cmd_IPACL_Std_Rule(_Cmd):
    context = "ip-acl-std"
    cmd = r"(?P<rule>(permit|deny) +.+)"

    def action(self, groups, cfg, params):
        cfg.append(groups["rule"])

_cmds.append(_Cmd_IPACL_Std_Rule)


class _Cmd_ACLExtRule(_Cmd):
    cmd = r"access-list (?P<num>1\d{2}|2[0-6]\d{2}) (?P<rule>.+)"

    def action(self, groups, cfg, params):
        deepsetdefault(cfg, "ip-acl-ext", groups["num"], last=[]).append(
            _ipacl_ext_rule_parse(groups["rule"]))

_cmds.append(_Cmd_ACLExtRule)


class _Cmd_IPACL_Ext(_Cmd):
    cmd = r"ip access-list extended (?P<name>.+)"
    enter_context = "ip-acl-ext"

    def action(self, groups, cfg, params):
        name = groups["name"]
        return deepsetdefault(cfg, "ip-acl-ext", name, last=[])

_cmds.append(_Cmd_IPACL_Ext)


class _Cmd_IPACL_Ext_Rule(_Cmd):
    context = "ip-acl-ext"
    cmd = r"(?P<rule>(permit|deny) +.+)"

    def action(self, groups, cfg, params):
        cfg.append(_ipacl_ext_rule_parse(groups["rule"]))

_cmds.append(_Cmd_IPACL_Ext_Rule)



# IPV6 ACCESS-LIST ...



class _Cmd_IP6ACL(_Cmd):
    cmd = r"ipv6 access-list (?P<name>.+)"
    enter_context = "ipv6-acl"

    def action(self, groups, cfg, params):
        name = groups["name"]
        return deepsetdefault(cfg, "ipv6-acl", name, last=[])

_cmds.append(_Cmd_IP6ACL)


class _Cmd_IP6ACL_Rule(_Cmd):
    context = "ipv6-acl"
    cmd = r"(?P<rule>(permit|deny) +.+)"

    def action(self, groups, cfg, params):
        cfg.append(_acl6_rule_parse(groups["rule"]))

_cmds.append(_Cmd_IP6ACL_Rule)



# IP[V6] PREFIX-LIST ...



class _Cmd_IPPfx(_Cmd):
    cmd = r"ip prefix-list (?P<list>\S+) (seq \d+ )?(?P<rule>.+)"

    def action(self, groups, cfg, params):
        list_ = groups["list"]

        deepsetdefault(cfg, "ip-pfx", list_, last=[]).append(groups["rule"])

_cmds.append(_Cmd_IPPfx)


class _Cmd_IP6Pfx(_Cmd):
    cmd = r"ipv6 prefix-list (?P<list>\S+) (seq \d+ )?(?P<rule>.+)"

    def action(self, groups, cfg, params):
        list_ = groups["list"]

        deepsetdefault(cfg, "ipv6-pfx", list_, last=[]).append(
            groups["rule"].lower())

_cmds.append(_Cmd_IP6Pfx)



# IP[V6] ROUTE ...



class _Cmd_IPRoute(_Cmd):
    cmd = r"ip route (?P<route>.+)"

    def action(self, groups, cfg, params):
        cfg.setdefault("ip-route", set()).add(groups["route"])

_cmds.append(_Cmd_IPRoute)


class _Cmd_IP6Route(_Cmd):
    cmd = r"ipv6 route (?P<route>.+)"

    def action(self, groups, cfg, params):
        # IPv6 addresses involve letters so we lower case for
        # consistency
        cfg.setdefault("ipv6-route", set()).add(groups["route"].lower())

_cmds.append(_Cmd_IP6Route)



# [NO] SPANNING-TREE ...



class _Cmd_NoSTP(_Cmd):
    cmd = r"no spanning-tree vlan (?P<tags>[-0-9,]+)"

    def action(self, groups, cfg, params):
        cfg.setdefault("no-stp", set()).update(_expand_set(groups["tags"]))

_cmds.append(_Cmd_NoSTP)


class _Cmd_STPPriority(_Cmd):
    cmd = r"spanning-tree vlan (?P<tags>[-0-9,]+) priority (?P<priority>\d+)"

    def action(self, groups, cfg, params):
        priority = groups["priority"]

        cfg_stp_pri = cfg.setdefault("stp-priority", {})
        for tag in _expand_set(groups["tags"]):
            cfg_stp_pri[int(tag)] = int(priority)

_cmds.append(_Cmd_STPPriority)



# VLAN ...



class _Cmd_VLAN(_Cmd):
    cmd = r"vlan (?P<tag>\d+)"
    enter_context = "vlan"

    def action(self, groups, cfg, params):
        tag = int(groups["tag"])

        # create the VLAN configuration entry, setting an 'exists' key
        # as we might stop other information in here that isn't in the
        # VLAN definition itself in IOS (e.g. STP priority) in future
        cfg_vlan = deepsetdefault(cfg, "vlan", tag)
        cfg_vlan["exists"] = None

        return cfg_vlan

_cmds.append(_Cmd_VLAN)


class _CmdContext_VLAN(_Cmd):
    context = "vlan"

class _Cmd_VLAN_Name(_CmdContext_VLAN):
    cmd = r"name (?P<name>\S+)"

    def action(self, groups, cfg, params):
        cfg["name"] = groups["name"]

_cmds.append(_Cmd_VLAN_Name)



# --- classes ----



class CiscoIOSConfig(IndentedContextualConfig):
    "This concrete class parses Cisco IOS configuration files."


    def _add_commands(self):
        """This method is called by the constructor to add commands for
        the IOS platform.

        The commands are stored in a global (to the module) level list
        of classes.
        """

        for cmd_class in _cmds:
            self._add_command(cmd_class)


    def _post_parse_file(self):
        """Extend the inherited method to flush any pending IPv4
        standard ACL rules into the configuration.
        """

        super()._post_parse_file()

        # go through the pending IPv4 standard ACLs and store them (we
        # convert this to a list as _acl4_std_flush() will change the
        # dictionary during iteration and we only need the names)
        if "ip-acl-std" in self:
            ip_acl_std = self["ip-acl-std"]
            for name in ip_acl_std:
                ip_acl_std[name] = _ipacl_std_canonicalize(ip_acl_std[name])



# configdiff.py
#
# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>



# --- imports ---



from copy import deepcopy
import re
import sys

from deepops import deepremoveitems, deepmerge, deepdiff, deepsetdefault, deepget
import yaml



# --- functions ---



def pathstr(path, wildcard_indices=set()):
    """This function converts a path, which is a list of items of
    various types (typically strings and integers), into a string,
    useful for debugging messages.

    The items are separated by commas and elements which are None are
    converted to a '*' since they're wildcards.

    If the index of the item in the path is in the wildcard_indices
    list, that item is prefixed with "*=" to show it matched a wildcard.
    The default for this parameter is an empty list, which causes no
    index to match.

    If we just printed the list, it would by in the more verbose Python
    format, with surrounding square brackets, quotes around strings and
    None for a wildcard, which is a little less readable.
    """

    return ":".join([ ("*=" if i in wildcard_indices else "") + str(v)
                            for i, v in enumerate(path) ])



# --- classes ---



class DiffConvert:
    """This abstract class handles converting the difference between a
    'from' configuration item to the corresponding 'to' configuration
    item.

    The main difference process will use deepops.deepdiff() to work out
    what has been removed and what updated (added/changed) between the
    two configurations.

    Individual differences are checked using child classes, which
    specify the part of the configuration directionaries where they
    occur and the remove() and update() methods called.  For example, if
    the hostname is changed, a 'hostname' converter would specify the
    part of the configuration dictionary where the hostname is stored
    and the update() method would return the commands required to change
    it to the new value.

    The child classes will override the following values:

    path -- a tuple giving the path of keys through the dictionaries to
    match for these actions; these can be literal keys or the value
    None, if that level of the path is to wildcard match all keys (for
    example, all interfaces); the part of the configuration dictionary
    that matches this path will be passed as the 'cfg' and 'diff'
    parameters to the action functions

    ext -- an extension to path which must be matched, but this part is
    not used to further specify the part of the dictionaries passed to
    the action functions for the 'cfg' and 'diff' parameters: this is
    useful when a higher level in the configuration dictionary is
    required in the action function

    name -- if there is more than one action acting at a particular path
    point, this can be defined to differentiate them: it's useful when
    using the insert_before parameter, but isn't needed if the order
    does not matter (except for ease of debugging, to clarify which
    converter is being called)

    insert_before -- if this rule needs to happen prior to the actions
    of another rule, this 2-tuple gives the (path, name) before the
    earlier rule; None means the order does not matter

    In addition the remove() and update() methods will likely need to be
    implemented.
    """


    path = None
    ext = tuple()
    name = None
    insert_before = None


    def __init__(self):
        """The constructor just precalculates some details to optimise
        the repeated processing of the converter matches.
        """

        # store the full path, including the extension
        self.path_full = self.path + self.ext

        # store the number of elements in the path
        self.path_len = len(self.path)

        # store the set of indices of wildcard elements of the path (we
        # need these to get the argument list to pass to the converter
        # action methods)
        self.wildcard_indices = {
            i for i, v in enumerate(self.path_full) if v is None }


    def pathmatches(self, d):
        """This method takes a dictionary and returns any matches for
        the path in it, as a list of paths; each path is, itself, a
        list of keys.

        If any of the elements of the path are None, this is treated as
        a wildcard and will match all keys at that level.

        If there are no matching entries, the returned list will be
        empty.

        The returned list of matches is not guaranteed to be in any
        particular order.
        """


        def _pathmatches(d, path):
            """This function is used to recursively step along the
            paths.
            """

            # if the path is finished, return a single result with an
            # empty list, as there are no more path elements
            #
            # note that this is different from an empty list (which
            # would mean no matches
            if not path:
                return [ [] ]

            # get the path head and tail to make things easier to read
            path_head, path_tail = path[0], path[1:]

            # if this element is not a type we can iterate or search
            # through (we've perhaps reached a value or a None), or the
            # dictionary is empty, there are no matches, so just return
            # an empty list (which will cause all higher levels in the
            # path that matched to be discarded, giving no resulsts)
            if not isinstance(d, (dict, list, set)) or (not d):
                return []

            # if the path element is None, we're doing a wildcard match
            if path_head is None:
                # initialise an empty list for all returned matching
                # results
                results = []

                # go through all the keys at this level in the dictonary
                for d_key in d:
                    # are there levels below this one?
                    if path_tail:
                        # yes - recursively get the matches from under
                        # this key
                        for matches in _pathmatches(d[d_key], path_tail):
                            # add this match to the list of matches,
                            # prepending this key onto the start of the
                            # path
                            results.append([d_key] + matches)
                    else:
                        # no - just add this result (this is a minor
                        # optimisation, as well as avoiding an error by
                        # trying to index into a non-dict type), above
                        results.append([d_key])

                return results

            # we have a literal key to match - if it's found,
            # recursively get the matches under this level
            if path_head in d:
                return [ ([path_head] + matches)
                             for matches
                             in _pathmatches(d[path_head], path_tail) ]

            # we have no matches, return the empty list
            return []

        # start mathcing with the full path
        return _pathmatches(d, self.path_full)


    def remove(self, cfg, diff, args):
        """The remove() method is called when the specified path is in
        the remove differences (i.e. it's in the 'from' configuration
        but not in the 'to' configuration).

        Keyword arguments:

        cfg -- the value of the dictionary item at the matching path in
        the 'from' configuration dictionary (sometimes, the full details
        of the old configuration may be required to remove it)

        diff -- the value of the dictionary item at the matching path in
        the remove differences dictionary (sometimes, it's necessary to
        know only what is removed from the new configuration)

        args -- [only] the arguments in the path which were matched with
        wildcards; if there were no wildcards, this list will be empty

        The return value is the commands to insert into the
        configuration to convert it.  This can either be a simple
        string, in the case of only one line being required, or an
        iterable containing one string per line.  If the return value is
        None, it is an indication nothing needed to be done.  An empty
        string or iterable indicates the update did something, which did
        not change the configuration (this may have semantic
        differences).
        """

        pass


    def update(self, cfg, diff, args):
        """The update() method is called when the specified path is in
        the update differences (i.e. it's in the 'to' configuration but
        not in the 'from' configuration, or the value differs between
        the two and needed to be changed).

        Keyword arguments:

        cfg -- the value of the dictionary item at the matching path in
        the 'to' configuration dictionary (sometimes, teh full details
        of the new configuration may be required to update it)

        diff -- the value of the dictionary item at the matching path in
        the update differences dictionary

        args -- [only] the arguments in the path which were matched with
        wildcards; if there were no wildcards, this list will be empty

        The return value is the commands to insert into the
        configuration to convert it.  This can either be a simple
        string, in the case of only one line being required, or an
        iterable containing one string per line.  If the return value is
        None, it is an indication nothing needed to be done.  An empty
        string or iterable indicates the update did something, which did
        not change the configuration (this may have semantic
        differences).
        """

        pass


    def get_cfg(self, cfg, match):
        """This method returns the configuration to be passed to the
        converter's action methods [remove() and update()].

        By default, it indexes through the configuration using the path
        in the converter.  Converters may wish to override this, in some
        cases, for performance (perhaps if the entire configuration is
        to be returned).

        If the specified match path is not found, or there was a problem
        indexing along the path, None is returned, rather than an
        exception raised.
        """

        # try to get the specified matching path
        try:
            return deepget(cfg, *match[0 : self.path_len])

        except TypeError:
            # we got a TypeError so make the assumption that we've hit
            # a non-indexable element (such as a set) as the final
            # element of the path, so just return None
            return None


    def get_ext(self, cfg):
        """This method gets the extension part of the path, given a
        configuration dictionary starting at the path (i.e. what is
        passed as 'cfg' and 'diff' in the action methods).

        An action method [remove() or update()] can use this to get the
        extension portion without needing to explicitly index through
        it.
        """

        return deepget(cfg, *self.ext)


    def get_wildcard_args(self, match):
        """This method returns a list of the wildcarded parts of the
        specified match as a list.  For example, if the path of this
        converter is 'interface', None) and the match is ['interface',
        'Vlan100'], the returned result will be ['Vlan100'].  This is
        used to supply the 'args' parameter to action methods.
        """

        return [ match[i] for i in self.wildcard_indices ]



class DiffConfig:
    """This abstract class is used to represent a configuration
    difference processor that can convert a configuration from one to
    another, using a method (which can be called once for each pair of
    configuration files).

    It encapsulates the rules for exluding items from the comparison.
    """


    def __init__(self, init_explain=False, init_debug_config=False,
                       init_debug_diff=False, init_debug_convert=0):

        """The constructor initialises the exclude list to empty and
        adds and converters using _add_converters().  It also stores
        some settings controlling the level of information describing
        the conversion process, based on the command line arguments:

        init_explain=False -- include comments in the output
        configuration changes that explain the differences being matched
        by the DiffConvert objects (if available).

        init_debug_config=False -- dump the 'from' and 'to'
        configurations (after exludes) to stderr

        init_debug_diff=False -- dump the differences (remove and update
        configurations) to stderr

        init_debug_convert=0 -- level of debugging information for the
        conversion process: >= 1: include steps, >= 2: include 'cfg' and
        'diff' parameters
        """

        # store the [initial] settings for the conversion process
        self._explain = init_explain
        self._debug_config = init_debug_config
        self._debug_diff = init_debug_diff
        self._debug_convert = init_debug_convert

        # initialise the dictionary of excludes, which will be passed
        # to deepremoveitems()
        self.init_excludes()

        # initialise the list of converters and add them
        self._cvtrs = []
        self._add_converters()


    def _add_converters(self):
        """This method adds the converters for a particular object to
        the list used by the convert() method, usually by calling
        _add_converter() for each (see that method).

        The base class does nothing but child classes will implement it
        as they require.
        """

        pass


    def _add_converter(self, cvtr):
        """Add an individual convert object, a child of the
        DiffConverter class, to the list of converters to be used by the
        convert() method.

        If the converter has an 'insert_before' entry that is not None,
        the converter will be inserted in the postition immediately
        before the existing entry with that name.  If that entry cannot
        be found, a KeyError will be raised.
        """

        if not cvtr.insert_before:
            # no 'insert_before' - just add it to the end
            self._cvtrs.append(cvtr)

        else:
            # 'insert_before' so we need to find that entry

            # get the indexes and entries in the converter list
            for pos, chk_cvtr in enumerate(self._cvtrs):
                # if this entry is the one we're inserting before, put
                # it here and stop
                if chk_cvtrs.name == cvtrs.insert_before:
                    self._cvtrs.insert(pos, cvtr)
                    break

            else:
                # we got the end of the list and didn't find it, so
                # raise an exception
                raise KeyError("cannot find existing converter with name: "
                               + cvtr.insert_before)


    def _explain_comment(self, path):
        """This method returns a comment or other configuration item
        explaining the path supplied (which will typically be a match
        against a converter).  The path is supplied a list of levels and
        converted to a string using pathstr().

        Child classes can override this to provide a comment appropriate
        for their platform.

        If the function returns None, no comment is inserted.
        """

        return None


    def _changes_begin(self):
        """This method returns a head (beginning) for a generated
        changes configuration file as an iterable of strings.

        In the abstract class, it returns None (which does not add
        anything) but, in child classes it may return a beginning line
        or similar.

        Note that if there are no changes, this will not be included in
        an otherwise empty file.
        """

        return []


    def _changes_end(self):
        """This method returns a tail (ending) for a generated changes
        configuration file as an iterable of strings.

        In the abstract class, it returns None (which does not add
        anything) but, in child classes it may return an 'end' line or
        similar.

        Note that if there are no changes, this will not be included in
        an otherwise empty file.
        """

        return []


    def init_excludes(self):
        """This method initialises the excludes dictionary.  In the base
        class, it the resets it to an empty dictionary, but child
        classes can extend this to add some default exclusions for
        standard system configuration entries which should not be
        changed.

        It is normally only called by the constructor but can be called
        later, to clear the excludes list before adding more (if
        required).
        """

        self._excludes = {}


    def add_exclude(self, path):
        """Add a single exclude item to the dictionary.  The exclude is
        supplied as a string, with the levels in the path of the
        configuration dictionary separated by colons (":"), and then
        added to the excludes dictionary.
        """

        # start at the top of the excludes dictionary
        excludes_sub = self._excludes

        # work through the levels in the exclude item path
        for level in path.split(":"):
            # create this level of the excludes dictionary as a new
            # empty dictionary and move down to it for the next level
            excludes_sub = excludes_sub.setdefault(level, {})


    def add_excludes(self, paths):
        """Add a list of exclude items as strings to the excludes, using
        add_exclude().
        """

        for path in paths:
            self.add_exclude(path)


    def read_excludes(self, filename):
        """Read a list of exclude items from a file and add them to the
        excludes dictionary using add_exclude().

        A hash ('#') and anything following it is stripped out as a
        comment.  Trailing whitespace is also removed.  Blank lines
        (including those created by the above actions) are skipped.
        """

        # work through the lines in the file
        for l in open(filename):
            # strip out anything following a '#' (indicating comment)
            # and any trailing whitespace
            l = re.sub("#.*", "", l)
            l = l.rstrip()

            # if the line is not blank, add it as an exclude item
            if l:
                self.add_exclude(l)


    def read_excludes_yaml(self, filename, path=None):
        """Read a list of exclude items from a YAML file and add them to
        the excludes dictionary.

        The path parameter specifies the key in the dictionary where the
        excludes are specified.  This is useful if the dictionary is
        also contains other data (such as the inventory).  The path is
        specified a list of keys, separated by colons.  If the path is
        None or otherwise False, the top of the dictionary is used.
        """


        # try to read in the YAML excludes file

        try:
            f = yaml.safe_load(open(filename))

        except yaml.parser.ParserError as exception:
            raise ValueError("failed parsing YAML excludes file: %s: %s"
                                 % (filename, exception))


        # if no path was specified, just use the entire file, else get
        # the key given

        if not path:
            deepmerge(self._excludes, f)
        else:
            deepmerge(self._excludes,
                      deepget(f, *path.split(":"), default_error=True))


    def get_excludes(self):
        "This method just returns the excludes dictionary."

        return self._excludes


    def convert(self, from_cfg, to_cfg):
        """This method processes the conversion from the 'from'
        configuration to the 'to' configuration, removing excluded parts
        of each and calling the applicable converters' action methods.

        Note that, if excludes are used, the configurations will be
        modified in place by a deepremoveitems().  They will need to be
        copy.deepcopy()ed before passing them in, if this is
        undesirable.

        The returned value is a single big string of all the
        configuration changes that need to be made.
        """


        # remove any excludes from the 'from' config

        deepremoveitems(from_cfg, self._excludes)

        if self._debug_config:
            print(">> 'from' configuration:",
                yaml.dump(from_cfg, default_flow_style=False), sep="\n",
                file=sys.stderr)


        # if no 'to' config was specified, stop here (we assume we're
        # just testing, parsing and excluding items from the 'from'
        # configuration and stopping

        if not to_cfg:
            return None


        # remove any excludes from the 'to' config

        deepremoveitems(to_cfg, self._excludes)

        if self._debug_config:
            print(">> 'to' configuration:",
                yaml.dump(to_cfg, default_flow_style=False), sep="\n",
                file=sys.stderr)


        # use deepdiff() to work out the differences between the two
        # configuration dictionaries - what must be removed and what
        # needs to be added or updated

        remove_cfg, update_cfg = deepdiff(from_cfg, to_cfg)

        if self._debug_diff:
            print("=> differences - remove:",
                  yaml.dump(remove_cfg, default_flow_style=False),
                  "=> differences - update (add/change):",
                  yaml.dump(update_cfg, default_flow_style=False),
                  sep="\n", file=sys.stderr)


        # initialise the list of changes (the returned configuration
        # conversions)

        changes = []


        # go through the list of converter objects

        for cvtr in self._cvtrs:
            if self._debug_convert:
                print(">> checking difference:",
                      pathstr([ "*" if i is None else i for i in cvtr.path ]),
                      "name:", cvtr.name, file=sys.stderr)


            # get all the remove and update matches for this converter
            # and combine them into one list, discarding any duplicates
            #
            # we do this rather than processing each list one after the
            # other so we combine removes and updates on the same part
            # of the configuration together

            remove_matches = cvtr.pathmatches(remove_cfg)
            update_matches = cvtr.pathmatches(update_cfg)

            all_matches = []
            for match in sorted(remove_matches + update_matches):
                if match not in all_matches:
                    all_matches.append(match)


            # go through all matches

            for match in all_matches:
                # handle REMOVE conversions, if matching

                if match in remove_matches:
                    if self._debug_convert:
                        print("=> remove match:",
                              pathstr(match, cvtr.wildcard_indices),
                              file=sys.stderr)


                    # get elements in the path matching wildcards

                    args = cvtr.get_wildcard_args(match)


                    # get the 'from' and 'remove' parts of the
                    # configuration and remove difference dictionaries,
                    # for the path specified in the converter (ignoring
                    # the extension 'ext')

                    cvtr_cfg = cvtr.get_cfg(from_cfg, match)
                    cvtr_diff = cvtr.get_cfg(remove_cfg, match)

                    if self._debug_convert >= 2:
                        print("-> from configuration:",
                            yaml.dump(
                                cvtr_cfg, default_flow_style=False),
                            "-> remove configuration:",
                            yaml.dump(
                                cvtr_diff, default_flow_style=False),
                            sep="\n", file=sys.stderr)


                    # call the remove converter action method using the
                    # corresponding entry from the 'from' configuration

                    change = cvtr.remove(cvtr_cfg, cvtr_diff, args)


                    # if some changes were returned by the action, add
                    # them

                    if change is not None:
                        # the return can be either a simple string or a
                        # list of strings - if it's a string, make it a
                        # list of one so we can do the rest the same way

                        if isinstance(change, str):
                            change = [change]

                        if self._debug_convert:
                            print("\n".join(change), file=sys.stderr)


                        # add a comment, explaining the match, if
                        # enabled

                        if self._explain:
                            comment = self._explain_comment(match)
                            if comment:
                                changes.append(comment)


                        # store this change on the end of the list of
                        # changes so far

                        changes.extend(change)
                        changes.append("")

                    else:
                        if self._debug_convert:
                            print("-> no action", file=sys.stderr)


                # handle UPDATE conversions, if matching

                if match in update_matches:
                    if self._debug_convert:
                        print("=> update match:",
                              pathstr(match, cvtr.wildcard_indices),
                              file=sys.stderr)


                    args = cvtr.get_wildcard_args(match)


                    # get the 'from' and 'update' parts of the
                    # configuration and update difference dictionaries,
                    # for the path specified in the converter (ignoring
                    # the extension 'ext')

                    cvtr_cfg = cvtr.get_cfg(to_cfg, match)
                    cvtr_diff = cvtr.get_cfg(update_cfg, match)

                    if self._debug_convert >= 2:
                        print("-> to configuration:",
                            yaml.dump(
                                cvtr_cfg, default_flow_style=False),
                            "-> update configuration:",
                            yaml.dump(
                                cvtr_diff, default_flow_style=False),
                            sep="\n", file=sys.stderr)


                    # call the update converter action method

                    change = cvtr.update(cvtr_cfg, cvtr_diff, args)


                    if change is not None:
                        if isinstance(change, str):
                            change = [change]

                        if self._debug_convert:
                            print("\n".join(change), file=sys.stderr)


                        if self._explain:
                            comment = self._explain_comment(match)
                            if comment:
                                changes.append(comment)


                        changes.extend(change)
                        changes.append("")

                    else:
                        if self._debug_convert:
                            print("-> no action", file=sys.stderr)


            # print a blank line if debugging, just to space things out
            # nicely

            if self._debug_convert:
                print(file=sys.stderr)

        # if nothing was generated, just return nothing

        if not changes:
            return None


        # return the change lines concatenated into a big, multiline
        # string, along with the begin and end blocks

        return "\n".join(self._changes_begin() + changes + self._changes_end())
